/* 
  Copyright (c) 2024, Anzo

  This program is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation, either version 3 of the License, or (at your option) any later
  version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
  details.
 
  You should have received a copy of the GNU General Public License along with 
	this program. If not, see <https://www.gnu.org/licenses/>.
																																						
===============================================================================
  Usage : %s [-o output file] [-n chunk size] [-t thread number] SOURCE TARGET

  This program Takes a target audio, splits it into chunks, and try to find the
  best subchunks in source that matches the target chunks (as in vector
  distance).
 
  SOURCE and TARGET should be mono wav file of sample rate 48000
																																						 */
#include "arg.h"
#include "tinywav.h"
#include <stdio.h>
#include <stdint.h>
#include <string.h> 
#include <stdlib.h>
#include <float.h>
#include <math.h>

#define NUM_CHANNELS 1
#define SAMPLE_RATE 48000
#define BLOCK_SIZE 480

#define DEBUG 1
#define PROCESS_REMAIN 1 //wether to process or not the trailing block with
												 //size < sample_amount

int sample_amount;
char* sourceFilename;
char* targetFilename;
char* outFilename;
TinyWav source;
TinyWav target;
TinyWav out;

/**
 *	Returns the best subsample (as in multiple sample points)
 *	of size len in source matching the target sample 
 *	(by index, the subsample being located at source+idx)
 *
 *	@param source			where to look for matching subsample (float*)
 *	@param target			what sample to match (float*)
 *	@param lenTarget	size of target, should be smaller than lenSource (int)
 *	@param lenSource	size of source (int)
 *
 *	@return the idx where to find the best subsample
 */
int matchingIdx(float* source, float* target, int lenTarget, int lenSource)
{
	int i = 0;
	int j = 0;
	int foundIdx = 0;
	float bestVal = 0.0;
	float val = 0.0;

	//needed for correct while loop condition and stuff. u smart.
	for(j = 0; j < lenTarget; j++)
	{bestVal += fabs(source[j] - target[j]);}
	
	while(i+lenTarget-1 < lenSource)
	{//lenTarget-th elements starting from i are inbound in source.
		
		val=0.0;
		//inner loop, we get the distance of the vectors source[i:n] and target 
		//as the sum of the abs of their differences
		for(j = 0; j < lenTarget; j++)
		{val += fabs(*(source+i+j) - target[j]);}
		
		if (val < bestVal)//min found, update value
		{foundIdx = i; bestVal = val;}
		i++;
	}
	if (DEBUG) {printf("%g %i", bestVal, foundIdx);}
	return foundIdx;
} 

int main(int argc, char* argv[])
{
	/*args parsing*/
	
	//default values
	outFilename = "out.wav";
	sample_amount = 1000;

	//arg.h
	ARGBEGIN {
		case 'o':
			outFilename = ARGF();
			break;
		case 'n':
			sample_amount = atoi(ARGF());
			break;
		default:
			fprintf(stderr,"unrecognized option : %s\n", ARGC());
			break;
  } ARGEND
	
	//retrieving source and target files
	if (argc != 2)
	{
		if (argc != 0) 
		{fprintf(stderr, "Expected two arguments (source target), %i were given\n"
				,argc);}

		fprintf(stderr, "Usage : %s [-o output file]\
[-n sample amount per chunk] SOURCE TARGET\n", argv0);
		exit(1);
	}
	sourceFilename = strdup(*argv);
	argv++;
	targetFilename = strdup(*argv);

	/*retrieving files*/
	tinywav_open_read(&source, sourceFilename, TW_INLINE);
	tinywav_open_read(&target, targetFilename, TW_INLINE);

	if (DEBUG)
	{
		printf("retrieved : %s %s, -n %i, -o %s\n"
				,sourceFilename
				,targetFilename
				,sample_amount
				,outFilename);
	}

	/*preparing output file*/
	tinywav_open_write(&out, NUM_CHANNELS, SAMPLE_RATE, TW_FLOAT32, TW_INLINE
			,outFilename);

	/*putting source and target file in buffers (fread)*/	
	int lenSource = source.numFramesInHeader;	
	float* sourceBuffer = (float *) malloc(sizeof(float)*lenSource);
	if (DEBUG) {printf("lenSource: %i\n", lenSource);} 
	tinywav_read_f(&source, sourceBuffer, lenSource);
	
	int lenTarget = target.numFramesInHeader;	
	float* targetBuffer = (float *) malloc(sizeof(float)*lenTarget); 
	if (DEBUG) {printf("lenTarget: %i\n", lenTarget);}
	tinywav_read_f(&target, targetBuffer, lenTarget);

	int i = 0;
	while (i < lenTarget)
	{

		int foundIdx = matchingIdx(sourceBuffer, targetBuffer+i
				, sample_amount, lenSource);
		tinywav_write_f(&out, sourceBuffer+foundIdx, sample_amount);

		i += sample_amount;
		
		printf("\n\033[1AProcessed %i / %i sample points..."
				, i, lenTarget);
		//final bit is smaller. can take much longer to process
		if (i + sample_amount >= lenTarget)
		{	
			if (PROCESS_REMAIN)
			{
				sample_amount = (lenTarget-i);
				int foundIdx = matchingIdx(sourceBuffer, targetBuffer+i
						, sample_amount, lenSource);
				tinywav_write_f(&out, sourceBuffer+foundIdx, sample_amount);
				
				printf("\n\033[1AProcessed %i / %i sample points..."
				, i-sample_amount, lenTarget);
			}
			break; 
		}

	} 
	
	tinywav_close_read(&source);
	tinywav_close_read(&target);
	tinywav_close_write(&out);	
	return 0;
} 
