/* 
  Copyright (c) 2024, Anzo

  This program is free software: you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation, either version 3 of the License, or (at your option) any later
  version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
  details.
 
  You should have received a copy of the GNU General Public License along with 
	this program. If not, see <https://www.gnu.org/licenses/>.
																																						
===============================================================================
  Usage : %s [-o output file] [-n chunk size] [-t thread number] SOURCE TARGET

  This program Takes a target audio, splits it into chunks, and try to find the
  best subchunks in source that matches the target chunks (as in vector
  distance).
 
  SOURCE and TARGET should be mono wav file of sample rate 48000

  This version uses pthread and has a -t option															 
																																						 */
#include "arg.h"
#include "tinywav.h"
#include <stdio.h>
#include <stdint.h>
#include <string.h> 
#include <stdlib.h>
#include <float.h>
#include <math.h>
#include <pthread.h>
#include <unistd.h>

#define NUM_CHANNELS 1
#define SAMPLE_RATE 48000

#define DEBUG 0
#define PROCESS_REMAIN 1 //wether to process or not the trailing block with
												 //size < sample_amount
int sample_amount;
int NUM_THREAD;
char* sourceFilename;
char* targetFilename;
char* outFilename;
TinyWav source;
TinyWav target;
TinyWav out;

/**
 *	Returns the best subsample (as in multiple sample points)
 *	of size len in source matching the target sample 
 *	(by index, the subsample being located at source+idx)
 *
 *	@param source			where to look for matching subsample (float*)
 *	@param target			what sample to match (float*)
 *	@param lenTarget	size of target, should be smaller than lenSource (int)
 *	@param lenSource	size of source (int)
 *
 *	@return the idx where to find the best subsample
 */
int matchingIdx(const float* source, const float* target
		, const int lenTarget, const int lenSource)
{
	int i = 0;
	int j = 0;
	int foundIdx = 0;
	float bestVal = 0.0;
	float val = 0.0;

	//needed for correct while loop condition and stuff. u smart.
	for(j = 0; j < lenTarget; j++)
	{bestVal += fabs(source[j] - target[j]);}
	
	while(i+lenTarget-1 < lenSource)
	{//lenTarget-th elements starting from i are inbound in source.
		
		val=0.0;
		//inner loop, we get the distance of the vectors source[i:n] and target 
		//as the sum of the abs of their differences
		for(j = 0; j < lenTarget; j++)
		{val += fabs(*(source+i+j) - target[j]);}
		
		if (val < bestVal)//min found, update value
		{foundIdx = i; bestVal = val;}
		i++;
	}
	//if (DEBUG) {printf("%g %i", bestVal, foundIdx);}
	return foundIdx;
} 

typedef struct 
{
	float* s;
	float* t;
	float* o;
	int lT;
	int lS;
	int i;
	int* a;
} thread_arg_t; 

void* thread_wrap(void *arg)
{
	int i = 0;
	thread_arg_t* data = (thread_arg_t*) arg;

	const int lenTarget = data->lT;
	const int lenSource = data->lS;
	const float* sourceBuffer = data->s;
	const float* targetBuffer = data->t;
	float* outBuffer = data->o;
	int* amount = data->a; 
	int id = data->i;

	while (i < lenTarget)
	{
		int foundIdx = matchingIdx(sourceBuffer, targetBuffer+i
				, sample_amount, lenSource);
		memcpy(outBuffer+i, sourceBuffer+foundIdx, sample_amount*sizeof(float));

		i += sample_amount;
		
		//printf("Processed %i / %i sample points (at %p)\n"
		//		, i, lenTarget, targetBuffer);
		if (i%2000 == 0) {*amount = i;}

		//final bit is smaller. can take much longer to process
		if (i + sample_amount >= lenTarget)
		{	
			if (PROCESS_REMAIN)
			{
				sample_amount = (lenTarget-i);
				int foundIdx = matchingIdx(sourceBuffer, targetBuffer+i
						,sample_amount, lenSource);
				memcpy(outBuffer+i, sourceBuffer+foundIdx
						, sample_amount*sizeof(float));
				*amount = lenTarget;	
				//printf("Processed %i / %i sample points (at %p)\n"
				//, i-sample_amount, lenTarget, targetBuffer);
			}
			break; 
		}
	} 
} 

void* report_progress(void* arg)
{
	thread_arg_t* data = (thread_arg_t*)arg;
	int val = 0;
	while(val < data->lT)
	{ 
		val = 0;
		for (int i=0; i < NUM_THREAD; i++)
		{val = val + *((data->a)+i); usleep(500*NUM_THREAD);} 
	printf("\n\033[1AProcessed %i / %i sample points ..."
		, val, data->lT);
	}
} 

int main(int argc, char* argv[])
{
	/*args parsing*/
	
	//default values
	outFilename = "out.wav";
	sample_amount = 1000;
	NUM_THREAD = 2;

	//arg.h
	ARGBEGIN {
		case 'o':
			outFilename = ARGF();
			break;
		case 'n':
			sample_amount = atoi(ARGF());
			break;
		case 't':
			NUM_THREAD = atoi(ARGF());
			break;
		default:
			fprintf(stderr,"unrecognized option : %s\n", ARGC());
			break;
  } ARGEND
	
	//retrieving source and target files
	if (argc != 2)
	{
		if (argc != 0) 
		{fprintf(stderr, "Expected two arguments (source target), %i were given\n"
				,argc);}

		fprintf(stderr, "Usage : %s [-o output file]\
[-n sample amount per chunk] [-t thread number] SOURCE TARGET\n", argv0);
		exit(1);
	}
	sourceFilename = strdup(*argv);
	argv++;
	targetFilename = strdup(*argv);

	/*retrieving files*/
	tinywav_open_read(&source, sourceFilename, TW_INLINE);
	tinywav_open_read(&target, targetFilename, TW_INLINE);

	if (DEBUG)
	{
		printf("retrieved : %s %s, -n %i, -o %s, -t %i\n"
				,sourceFilename
				,targetFilename
				,sample_amount
				,outFilename
				,NUM_THREAD);
	}

	/*preparing output file*/
	tinywav_open_write(&out, NUM_CHANNELS, SAMPLE_RATE, TW_FLOAT32, TW_INLINE
			,outFilename);

	/*putting source and target file in buffers (fread)*/	
	int lenSource = source.numFramesInHeader;	
	float* sourceBuffer = (float *) malloc(sizeof(float)*lenSource);
	if (DEBUG) {printf("lenSource: %i\n", lenSource);} 
	tinywav_read_f(&source, sourceBuffer, lenSource);
	
	int lenTarget = target.numFramesInHeader;	
	float* targetBuffer = (float *) malloc(sizeof(float)*lenTarget); 
	if (DEBUG) {printf("lenTarget: %i\n", lenTarget);}
	tinywav_read_f(&target, targetBuffer, lenTarget);
	
	pthread_t threads[NUM_THREAD+1];
	thread_arg_t thread_arg[NUM_THREAD+1];
	
	int tsize = lenTarget/NUM_THREAD;
	int residue = lenTarget % NUM_THREAD;
	
	//printf("\033[2J");
	
	int proginfo[NUM_THREAD];//to get progression info from threads.
	memset(proginfo, NUM_THREAD*sizeof(int), 0);

	for (int i=0; i < NUM_THREAD; i++)
	{
		thread_arg[i].i = i;
		thread_arg[i].lT = tsize + (i == (NUM_THREAD-1) ? residue : 0);
		thread_arg[i].lS = lenSource;
		thread_arg[i].s = sourceBuffer;
		thread_arg[i].t = targetBuffer+(i*tsize);
		thread_arg[i].o = (float*) malloc(sizeof(float)*thread_arg[i].lT); 
		thread_arg[i].a = &proginfo[i];  

		pthread_create(&threads[i], NULL, &thread_wrap ,&thread_arg[i]);
	}
	
	thread_arg[NUM_THREAD].a = proginfo;
	thread_arg[NUM_THREAD].lT = lenTarget;	
	pthread_create(&threads[NUM_THREAD], NULL, &report_progress
			, &thread_arg[NUM_THREAD]);

	for(int i=0; i < NUM_THREAD+1; i++)
	{
		pthread_join(threads[i], NULL);
	}

	//writing results.
	for (int i=0; i < NUM_THREAD; i++)
	{
		tinywav_write_f(&out, thread_arg[i].o, thread_arg[i].lT);
		free(thread_arg[i].o);
	} 

	free(targetBuffer);
	free(sourceBuffer);

	tinywav_close_read(&source);
	tinywav_close_read(&target);
	tinywav_close_write(&out);	
	return 0;
} 
